#include "common.h"
#include "scanner.h"
#include "expr.h"
#include "stmt.h"
#include "func.h"
#include "program.h"
#include "parser.h"
#include <stdio.h>
#include <string.h>

//forward definition
static void match(int expected);
static program_t *parse_program();
static list_t *parse_opt_var_list();
static list_t *parse_func_list();

static stmt_t *parse_stmt();
static stmt_t *parse_stmt_assignment();
static stmt_t *parse_stmt_read();
static stmt_t *parse_stmt_write();
static stmt_t *parse_stmt_compound();
static list_t *parse_stmt_list();

static expr_t *parse_expr();
static expr_t *parse_expr_add();
static expr_t *parse_expr_mul();
static expr_t *parse_expr_factor();

static stmt_t *parse_stmt_if();
static stmt_t *parse_stmt_while();
static expr_t *parse_expr_comp();

static func_t *parse_func();
static list_t *parse_name_list();
static stmt_t *parse_stmt_return();
static expr_t *parse_expr_fun_app();
static list_t *parse_expr_list();

//scanner state
static scan_t sc;

static void match(int expected) {
    ON_FALSE_EXIT( sc.token == expected,
                   strmsg("syntax error: in %d, expected %d, but has %d (%s)", sc.line, expected, sc.token, sc.text));
    read_token(&sc);
}

program_t *parse(char *fname) {
    open_scan(&sc, fname);
    read_token(&sc);
    program_t *pgm = parse_program();
    close_scan(&sc);
    return pgm;
}

//
// program
//
static program_t *parse_program() {
    list_t *vars  = parse_opt_var_list();
    list_t *funcs = parse_func_list();
    match(TOK_EOF);
    program_t *pgm = program_make(vars, funcs);
    for(list_t *pos = funcs->next; pos != funcs; pos = pos->next) {
        func_list_entry_t *entry = containerof(pos, func_list_entry_t, lst);
        entry->func->ref.release(&entry->func->ref);
    }
    return pgm;
}

static list_t *parse_opt_var_list() {
    if(sc.token == TOK_VAR) {
        match(TOK_VAR);
        return parse_name_list();
    }
    else {
        list_t *head = malloc(sizeof(list_t));
        list_init_head(head);
        return head;
    }
}

static list_t *parse_func_list() {
    list_t *head = malloc(sizeof(list_t));
    list_init_head(head);

    while(sc.token != TOK_EOF) {
        func_t *func = parse_func();
        func_list_entry_t *entry = malloc(sizeof(func_list_entry_t));
        entry->func = func;
        list_add_to_last(head, &entry->lst);
    }

    return head;
}

//
// function definitions
//
static func_t *parse_func() {
    match(TOK_FUN);
    char *fname = strdup(sc.text); //get var name from the scanner text
    match(TOK_ID);
    match('(');
    list_t *id_list = parse_name_list();
    match(')');
    stmt_t *body = parse_stmt_compound();

    func_t *func = func_make(fname, id_list, body);

    free(fname);
    body->ref.release(&body->ref);
    return func;
}

static list_t *parse_name_list() {
    list_t *head = malloc(sizeof(list_t));
    list_init_head(head);

    while(sc.token == TOK_ID) {
        name_list_entry_t *entry = malloc(sizeof(name_list_entry_t));
        entry->name = strdup(sc.text); //get var name from the scanner text
        match(TOK_ID);
        list_add_to_last(head, &entry->lst);
        if(sc.token != ',')
            break;
        match(',');
    }

    return head;
}

//
// statements
//
static stmt_t *parse_stmt() {
    stmt_t *stmt = NULL;
    if(sc.token == TOK_ID)
        stmt = parse_stmt_assignment();
    else if(sc.token == TOK_READ)
        stmt = parse_stmt_read();
    else if(sc.token == TOK_WRITE)
        stmt = parse_stmt_write();
    else if(sc.token == '{')
        stmt = parse_stmt_compound();
    else if(sc.token == TOK_IF)
        stmt = parse_stmt_if();
    else if(sc.token == TOK_WHILE)
        stmt = parse_stmt_while();
    else if(sc.token == TOK_RETURN)
        stmt = parse_stmt_return();
    else
        ON_FALSE_EXIT(0, strmsg("syntax error: in %d near (%s), not a statement", sc.line, sc.text));
    return stmt;
}

static stmt_t *parse_stmt_assignment() {
    char *id = strdup(sc.text);
    match(TOK_ID);
    match(TOK_ASSIGN);
    expr_t *expr = parse_expr();

    stmt_t *stmt = stmt_make_assignment(id, expr);

    expr->ref.release(&expr->ref);
    free(id);
    return stmt;
}

static stmt_t *parse_stmt_read() {
    match(TOK_READ);
    char *id = strdup(sc.text);
    match(TOK_ID);

    stmt_t *stmt = stmt_make_read(id);

    free(id);
    return stmt;
}

static stmt_t *parse_stmt_write() {
    match(TOK_WRITE);
    expr_t *expr = parse_expr();
    stmt_t *stmt = stmt_make_write(expr);

    expr->ref.release(&expr->ref);
    return stmt;
}

static stmt_t *parse_stmt_compound() {
    match('{');
    list_t *stmt_list = parse_stmt_list();
    match('}');

    stmt_t *stmt = stmt_make_compound(stmt_list);

    //release statements,
    for(list_t *pos = stmt_list->next; pos != stmt_list; pos = pos->next) {
        stmt_list_entry_t *entry = containerof(pos, stmt_list_entry_t, lst);
        entry->stmt->ref.release(&entry->stmt->ref);
    }

    return stmt;
}

static list_t *parse_stmt_list() {
    list_t *head = malloc(sizeof(list_t));
    list_init_head(head);

    while(sc.token != '}') {
        stmt_t *stmt = parse_stmt();
        stmt_list_entry_t *entry = malloc(sizeof(stmt_list_entry_t));
        entry->stmt = stmt;
        list_add_to_last(head, &entry->lst);
    }

    return head;
}

static stmt_t *parse_stmt_if() {
    match(TOK_IF);
    match('(');
    expr_t *cond = parse_expr();
    match(')');
    stmt_t *then_stmt = parse_stmt();
    match(TOK_ELSE);
    stmt_t *else_stmt = parse_stmt();

    stmt_t *stmt = stmt_make_if(cond, then_stmt, else_stmt);
    cond->ref.release(&cond->ref);
    then_stmt->ref.release(&then_stmt->ref);
    else_stmt->ref.release(&else_stmt->ref);
    return stmt;

}

static stmt_t *parse_stmt_while() {
    match(TOK_WHILE);
    match('(');
    expr_t *cond = parse_expr();
    match(')');
    stmt_t *loop_stmt = parse_stmt();

    stmt_t *stmt = stmt_make_while(cond, loop_stmt);

    cond->ref.release(&cond->ref);
    loop_stmt->ref.release(&loop_stmt->ref);
    return stmt;
}

static stmt_t *parse_stmt_return() {
    match(TOK_RETURN);
    expr_t *expr = parse_expr();
    stmt_t *stmt = stmt_make_return(expr);

    expr->ref.release(&expr->ref);
    return stmt;
}


//
// expressions
//
static expr_t *parse_expr() {
    return parse_expr_comp();
}

static expr_t *parse_expr_comp() {
    expr_t *op1 = parse_expr_add();
    expr_t *ret = op1;
    if( sc.token == TOK_EQ || sc.token == TOK_NE ||
        sc.token == TOK_LE || sc.token == '<' ||
        sc.token == TOK_GE || sc.token == '>' ) {
        int tok = sc.token;
        match(sc.token);
        expr_t *op2 = parse_expr_add();

        if(tok == TOK_EQ)
            ret = expr_make_eq(op1, op2);
        else if(tok == TOK_NE)
            ret = expr_make_ne(op1, op2);
        else if(tok == TOK_LE)
            ret = expr_make_le(op1, op2);
        else if(tok == '<')
            ret = expr_make_lt(op1, op2);
        else if(tok == TOK_GE)
            ret = expr_make_ge(op1, op2);
        else if(tok == '>')
            ret = expr_make_gt(op1, op2);

        op1->ref.release(&op1->ref);
        op2->ref.release(&op2->ref);
    }
    return ret;
}

static expr_t *parse_expr_add() {
    expr_t *op1 = parse_expr_mul();
    expr_t *ret = op1;
    while(sc.token == '+' || sc.token == '-') {
        int tok = sc.token;
        match(sc.token);
        expr_t *op2 = parse_expr_mul();
        
        expr_t *tmp;
        if(tok == '+')
            tmp = expr_make_add(op1, op2);
        else
            tmp = expr_make_sub(op1, op2);

        op1->ref.release(&op1->ref);
        op2->ref.release(&op2->ref);

        ret = op1 = tmp;    //op1 is the resulting expr
    }
    return ret;
}

static expr_t * parse_expr_mul() {
    expr_t *op1 = parse_expr_fun_app();
    expr_t *ret = op1;
    while(sc.token == '*' || sc.token == '/') {
        int tok = sc.token;
        match(sc.token);
        expr_t *op2 = parse_expr_fun_app();
        
        expr_t *tmp;
        if(tok == '*')
            tmp = expr_make_mul(op1, op2);
        else
            tmp = expr_make_div(op1, op2);

        op1->ref.release(&op1->ref);
        op2->ref.release(&op2->ref);

        ret = op1 = tmp;    //op1 is the resulting expr
    }
    return ret;
}

static expr_t *parse_expr_fun_app() {
    //save ID to make funciton
    char *id = (sc.token == TOK_ID ? strdup(sc.text): NULL);

    expr_t *opr = parse_expr_factor();
    expr_t *ret = opr;
    while(sc.token == '(') {
        match('(');
        list_t *arg_list = parse_expr_list();
        match(')');

        expr_t *tmp = expr_make_fun_app(id, arg_list);

        opr->ref.release(&opr->ref); //we are using fname instead of opr
        for(list_t *pos = arg_list->next; pos != arg_list; pos = pos->next) {
            expr_list_entry_t *entry = containerof(pos, expr_list_entry_t, lst);
            entry->expr->ref.release(&entry->expr->ref);
        }

        ret = tmp;
    }
    
    if(id)
        free(id);
    return ret;
}

static list_t *parse_expr_list() {
    list_t *head = malloc(sizeof(list_t));
    list_init_head(head);

    while(sc.token != ')') {
        expr_t *expr = parse_expr();
        expr_list_entry_t *entry = malloc(sizeof(expr_list_entry_t));
        entry->expr = expr;
        list_add_to_last(head, &entry->lst);
        if(sc.token != ',')
            break;
        match(',');
    }

    return head;
}

static expr_t *parse_expr_factor() {
    expr_t *ret = NULL;
    if(sc.token == TOK_NUM) {
        int n = atoi(sc.text);      //get int value from the scanner text
        match(TOK_NUM);
        ret = expr_make_num(n);
    }
    else if(sc.token == TOK_ID) {
        char *id = strdup(sc.text); //get var name from the scanner text
        match(TOK_ID);
        ret = expr_make_var(id);    //variable
        free(id);
    }
    else if(sc.token == '(') {
        match('(');
        ret = parse_expr();
        match(')');
    }
    return ret;
}

